from ursina import *
from ursina.prefabs.first_person_controller import FirstPersonController

app = Ursina()
grass_texture = load_texture('assets/grass_block.png')
stone_texture = load_texture('assets/stone_block.png')
brick_texture = load_texture('assets/brick_block.png')
dirt_texture  = load_texture('assets/dirt_block.png')
sky_texture   = load_texture('assets/skybox.png')
arm_texture   = load_texture('assets/arm_texture.png')
water_texture = load_texture('assets/water_texture.png')
punch_sound   = Audio('assets/punch_sound',loop = False, autoplay = False)
texture_bloc = grass_texture # texture d'herbe par défaut
BLOCS = dict() # dictionnaire des blocs, afin de pouvoir en détruire

window.fps_counter.enabled = False
window.exit_button.visible = False
# enlever les # dans les deux lignes suivantes pour activer du brouillard
# scene.fog_color = color.gray
# scene.fog_density = .02

# dans Ursina, la fonction update est exécutée régulièrement (après un temps time.dt) et sans arrêt, tant que le monde existe.
def update() :
    global texture_bloc
    if held_keys['left mouse'] or held_keys['right mouse'] :
        hand.active()
    else :
        hand.passive()
    if held_keys['h'] :
        texture_bloc = grass_texture # herbe
    if held_keys['p'] :
        texture_bloc = stone_texture # pierre
    if held_keys['b'] :
        texture_bloc = brick_texture # brique
    if held_keys['d'] :
        texture_bloc = dirt_texture # poussière (dirt)
    if held_keys['e'] :
        texture_bloc = water_texture # eau
    if held_keys['u'] :
        camera.position += (0, 5*time.dt, 0)# monter
    if held_keys['j'] :
        camera.position -= (0, 5*time.dt, 0)# descendre
    if mouse.hovered_entity == soleil : # affiche de l'aide lors du survol du soleil
        info.visible = True
    else :
        info.visible = False
    
# fonction reliée à Ursina qui permet de définir des interactions via le clavier avec le joueur
def input(key):
    if key == 'escape' or key == 'q' :
        quit()
    if key == 'm' : # touche "magique"
        magique()
      
# création d'une classe d'objets pour les blocs, basée sur la classe "Button" d'Ursina
# (la programmation orientée objet est au programme de Terminale)
class Bloc(Button) :
    def __init__(self, position = (0, 0, 0), texture = texture_bloc, taille = 0.5, couleur = color.color(0, 0, random.uniform(0.9, 1))):
        # par défaut, un bloc d'herbe de 0.5 de côté avec une teinte de vert plus ou moins prononcée
        Button.__init__(
            self,
            parent = scene,
            position = position,
            model = 'assets/block',
            origin_y = 0.5,
            texture = texture,
            color =  couleur,
            scale = taille
            )

    def input(self,key):
        if self.hovered: # si le bloc est survolé par la souris
            if key == 'left mouse down' : # et si l'on clique sur le bouton gauche
                punch_sound.play()
                Bloc(position = self.position + mouse.normal, texture = texture_bloc)
            if key == 'right mouse down' :
                punch_sound.play()
                destroy(self)

# et voici une main (ou un bras...) simpliste
class Hand(Entity) :
    def __init__(self) :
        super().__init__(
            parent = camera.ui,
            model = 'assets/arm',
            texture = arm_texture,
            scale = 0.2,
            rotation = Vec3(150,-10,0),
            position = Vec2(0.4,-0.6))

    def active(self) :
        self.position = Vec2(0.3,-0.5)

    def passive(self) :
        self.position = Vec2(0.4,-0.6)


# création d'une zone de texte d'aide (qui s'affichera quand on survole le soleil)
def creer_aide() :
    global soleil, info
    descr = dedent('''
          <scale:1.5><orange>Déplacements<default><scale:1>
          <orange>Flèches directionnelles du clavier<default> : se déplacer
          <orange>Souris<default> : s'orienter
          <orange>Barre d'espace<default> : pour sauter
          <orange>u<default> : pour voler (up)
          <orange>j<default> : pour descendre
          
          <scale:1.5><orange>Construire/Détruire<default><scale:1>
          <orange>Clic-gauche<default> : construire, <orange>clic-droit<default> : détruire
          Certaines touches permettent de modifier le type du bloc à construire :
          <orange>h<default> : herbe
          <orange>p<default> : pierre
          <orange>b<default> : brique
          <orange>d<default> : poussière (dirt)
          <orange>e<default> : eau
          
          <scale:1.5><orange>Touche magique<default><scale:1>
          <orange>m<default> : pour exécuter la fonction <it>magique
          ''').strip()
    Text.size = 0.02
    Text.default_resolution = 1080 * Text.size
    info = Text(text=descr, position = (-0.8, 0.4), background=True)
    info.visible = False # l'aide est cachée par défaut
    soleil = Entity(model = 'sphere', scale = 10, position = (LARG, 20, LONG), collider = 'box', color = color.yellow)

# par défaut, les mouvements se font avec les touches WASD
# ce qui n'est pas adapté aux claviers français
# la fonction suivante permet l'utilisation des flèches au lieu de WASD
def modifier_touches() :
    input_handler.rebind('left arrow', 'a')
    input_handler.rebind('right arrow', 'd')
    input_handler.rebind('up arrow', 'w')
    input_handler.rebind('down arrow', 's')
    input_handler.rebind('w', '&')
    input_handler.rebind('a', '&')
    input_handler.rebind('s', '&')
    input_handler.rebind('d', '&')

# fonction pour délimiter la zone et ne pas tomber dans le vide
def sol_et_murs() :
    # construction du sol
    # sous-sol (indestructible)
    sol = Entity(
        model = 'assets/block',
        color = color.gray,
        scale = (LARG//2, 0.5, LONG//2),
        #position = (LARG//2-0.5, -2, LONG//2-0.5),
        #origin = Vec3(0,0,0),
        position = Vec3(LARG//2-0.5, -2, LONG//2-0.5),
        collider = 'box'
        )

    # sol en "terre" (destructible)
    for z in range(LONG):
        for x in range(LARG):
            Bloc(position = (x, -1, z))

    # construction des murs
    mur_1 = Entity(model='cube', collider='box', scale = (LARG, 5, 0.5), x=LARG//2-0.5, y = 0.5, z = -0.75, color = color.rgba(0, 255, 0, 20))
    mur_2 = duplicate(mur_1, z = LONG-0.25)
    mur_3 = Entity(model='cube', collider='box', scale = (0.5, 5, LONG), z=LONG//2-0.5, x = -0.75, y = 0.5, color = color.rgba(0, 255, 0, 20))
    mur_4 = duplicate(mur_3, x = LARG-0.25)

# adaptation d'un code trouvé ici : https://github.com/fogleman/Minecraft/blob/master/main.py
def construire_collines(nb) :
        # generate the hills randomly
        for _ in range(nb) :
            s = random.randint(4, 8)  # 2 * s is the side length of the hill
            a = random.randint(s, LARG-s)  # x position of the hill
            b = random.randint(s, LONG-s)  # z position of the hill
            h = random.randint(1, 6)  # height of the hill
            d = 1  # how quickly to taper off the hills
            t = random.choice([grass_texture, stone_texture, brick_texture, dirt_texture]) # texture aléatoire
            
            for y in range(h) :
                for x in range(a - s, a + s + 1) :
                    for z in range(b - s, b + s + 1) :
                        if (x - a) ** 2 + (z - b) ** 2 > (s + 1) ** 2 :
                            continue
                        if x ** 2 + z ** 2 < 5 ** 2 :
                            continue
                        Bloc(position = (x, y, z), texture = t)
                s -= d  # decrement side length so hills taper off

def construire_bloc(x, y, z, cote = 1, couleur = color.color(0, 0, random.uniform(0.9, 1)), texture = texture_bloc) :
    cote = cote * 0.5 # l'unité est, dans ce fichier, de 0.5
    base = 1.5*cote - 0.75 # permet de placer le bloc au sol
    BLOCS[(x,y,z)] = Bloc(position = (x, base + z, y), texture = texture, taille = cote, couleur = couleur)

def detruire_bloc(x, y, z) :
    '''
    cette fonction permet de détruire le bloc placé aux coordonnées (x,y,z)
    '''
    if (x,y,z) in BLOCS :
        destroy(BLOCS[(x,y,z)])
        del(BLOCS[(x,y,z)])

def creer_monde(longueur = 50, largeur = 50) : # par défaut, les dimensions sont 50x50
    global sky, player, hand, LONG, LARG
    LONG = longueur # stockage des valeurs choisies par l'utilisateur dans des variables globales donc accessibles dans toutes les fonctions
    LARG = largeur
    # ajout des différents objets présents dans le monde
    sky = Sky()
    sol_et_murs()
    creer_aide()
    # la commande suivante lance la construction de 3 collines :
    #construire_collines(3)
    player = FirstPersonController(position = (2, 1, 2), speed = 5)
    player.gravity = 0.5
    hand = Hand()
    modifier_touches()
    app.run()

# fonction liée à la touche magique "m"
def magique() :
    pass    

creer_monde(40, 40)
